package org.expeditee.auth;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.KeyFactory;
import java.security.KeyStore;
import java.security.KeyStore.SecretKeyEntry;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.UnrecoverableEntryException;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.X509EncodedKeySpec;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Base64;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;
import java.util.stream.Stream;

import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.expeditee.actions.Actions;
import org.expeditee.auth.tags.Constants;
import org.expeditee.core.Dimension;
import org.expeditee.core.Point;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.gio.InputManager;
import org.expeditee.gio.InputManager.WindowEventListener;
import org.expeditee.gio.InputManager.WindowEventType;
import org.expeditee.gio.gesture.StandardGestureActions;
import org.expeditee.gui.Browser;
import org.expeditee.gui.DisplayController;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.FrameUtils;
import org.expeditee.gui.MessageBay;
import org.expeditee.io.ExpReader;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Text;
import org.expeditee.settings.Settings;
import org.expeditee.settings.UserSettings;
import org.expeditee.settings.identity.secrets.KeyList;
import org.ngikm.cryptography.CryptographyConstants;

public final class Authenticator implements CryptographyConstants {
	
	// The frame number of the frame containing the current authenticated users public key.
	public static int CREDENTIALS_FRAME = 13;
	
	public static boolean Authenticated = false;
	
	private final KeyStore keyStore = KeyStore.getInstance(KeystoreType);
		
	private static final byte[] TRUE = "yes".getBytes();
	private static final byte[] FALSE = "no".getBytes();
	private static final String KEYSTOREFILENAME = "keystore.ks" + File.separator;
	
	private static Authenticator instance;
	
	public static Authenticator getInstance() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, IOException, ClassNotFoundException, SQLException {
		if (instance == null) { instance = new Authenticator(); } 
		return instance;
	}
	
	private Authenticator() throws KeyStoreException, FileNotFoundException, IOException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, SQLException {
		UserSettings.setupDefaultFolders();
		
		// initialise keystore and actions
		loadKeystore();
		Actions.LoadMethods(org.expeditee.auth.Actions.class);

		// draw the window
		GraphicsManager g = EcosystemManager.getGraphicsManager();
		g.setWindowLocation(new Point(50, 50));
		DisplayController.Init();
		g.setWindowSize(new Dimension(UserSettings.InitialWidth.get(), UserSettings.InitialHeight.get()));
		setInputManagerWindowRoutines();
		
		// Load documentation and start pages
		FrameUtils.extractResources(false);

		// Load fonts before loading any frames so the items on the frames will be able to access their fonts
		Text.InitFonts();
		
		// initialing settings does not require a user profile established
		Settings.Init();
		
		// navigate to authentication frame
		Frame authFrame = FrameIO.LoadFrame("authentication1");
		DisplayController.setCurrentFrame(authFrame, true);
		
		// set initial values
		Stream<Text> usernameItemsStream = authFrame.getTextItems().stream().filter(t -> t.getData() != null && t.getData().contains("txtUsername"));
		Stream<Text> passwordItemsStream = authFrame.getTextItems().stream().filter(t -> t.getData() != null && t.getData().contains("txtPassword"));
		usernameItemsStream.forEach(txtUsername -> txtUsername.setText(System.getProperty("user.name", "")));
		passwordItemsStream.forEach(txtPassword -> txtPassword.setText(System.getProperty("user.password", "")));
						
		MessageBay.warningMessages(org.expeditee.actions.Actions.Init());
		
		// class load database classes
		Class.forName("org.sqlite.JDBC");
	}
		
	private void loadKeystore()
			throws IOException, NoSuchAlgorithmException, CertificateException, FileNotFoundException {
		final File keyStoreFile = new File(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME);
		if (!keyStoreFile.exists()) { 
			keyStore.load(null, Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
		} else {
			try (final InputStream in = new FileInputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME)) {
				keyStore.load(in, Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
			}
		}
	}
	
	final void loadMailFromDirectory(Path contactDir) throws SQLException {
		// Load in all mail.
		Connection c = DriverManager.getConnection("jdbc:sqlite:" + contactDir.resolve("expmail.db"));
		String sql = "SELECT * FROM EXPMAIL";
		PreparedStatement query = c.prepareStatement(sql);
		ResultSet allMail = query.executeQuery();

		// Construct all mail objects using content from database.
		while(allMail.next()) {
			String timestamp = allMail.getString("time");
			String from = allMail.getString("snd");
			String to = allMail.getString("rec");
			String msg = allMail.getString("msg");
			String msg2 = allMail.getString("msg2");
			String[] opts = allMail.getString("opts").split(",");
			opts[0] = opts[0].replace("[", "");
			opts[opts.length - 1] = opts[opts.length - 1].replace("]", "");
			String[] optsVal = allMail.getString("optsval").split(",");
			optsVal[0] = optsVal[0].replace("[", "");
			optsVal[optsVal.length - 1] = optsVal[optsVal.length - 1].replace("]", "");
			
			Map<String, String> options = new HashMap<String, String>();
			for (int i = 0, o = 0; i < opts.length && o < optsVal.length; i++, o++) {
				String key = opts[i].trim();
				String val = optsVal[o].trim();
				options.put(key, val);
			}
			
			Mail.addEntry(new Mail.MailEntry(timestamp, from, to, msg, msg2, options));
		}
		
		// Disconnect from database.
		allMail.close();
		query.close();
		c.close();
	}
	
	final void loadMailDatabase() throws SQLException {
		Path contactsPath = Paths.get(FrameIO.CONTACTS_PATH);
		File[] contacts = contactsPath.toFile().listFiles();
		for (int i = 0; i < contacts.length; i++) {
			if (contacts[i].isDirectory()) {
				Path contact = Paths.get(contacts[i].getAbsolutePath());
				loadMailFromDirectory(contact);
			}
		}
	}
	
	final SecretKey getSecretKey(final String label, final String password) throws NoSuchAlgorithmException, KeyStoreException {		
		final KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(password.toCharArray());
		KeyStore.SecretKeyEntry entry = null;
		try {
			entry = (SecretKeyEntry) keyStore.getEntry(label, entryPassword);
		} catch (final UnrecoverableEntryException e) {
			entry = null;
		}
		if (entry == null) { return null; }
		else { return entry.getSecretKey(); }
	}
	
	final void putKey(final String label, final String password, final SecretKey key) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException {
		final KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(key);
		final KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(password.toCharArray());
		keyStore.setEntry(label, entry, entryPassword);
		keyStore.store(new FileOutputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME), Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
	}
	
	final boolean confirmIntergalaticNumber(final String username, final String email, final String intergalacticNumber) throws NoSuchAlgorithmException, KeyStoreException, CertificateException, FileNotFoundException, IOException {
		try {
			final KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(intergalacticNumber.toCharArray());
			final KeyStore.SecretKeyEntry entry = (SecretKeyEntry) keyStore.getEntry(email + username, entryPassword);
			if (entry == null) { 
				return false;
			} else if (entry.getSecretKey().getEncoded() == TRUE) {
				keyStore.deleteEntry(email + username);
				keyStore.store(new FileOutputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME), Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
				return true;
			} else { return false; }
		} catch (final UnrecoverableEntryException e) {
			return false;
		}
	}
	
	final String newIntergalacticNumber(final String username, final String email) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException {
		// generate intergalactic number
		final SecureRandom rand = new SecureRandom();
		final byte[] intergalacticNumberBytes = new byte[20];
		rand.nextBytes(intergalacticNumberBytes);
		final String intergalacticNumber = Base64.getEncoder().encodeToString(intergalacticNumberBytes);
		
		// store intergalactic number
		final KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(new SecretKeySpec(TRUE, SymmetricAlgorithm));
		final KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(intergalacticNumber.toCharArray());
		keyStore.setEntry(email + username, entry, entryPassword);
		keyStore.store(new FileOutputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME), Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
		
		return intergalacticNumber;
	}
	
	final PublicKey getPublicKey(String username) throws InvalidKeySpecException, NoSuchAlgorithmException, FileNotFoundException {
		// load in frame with public key on it.
		String credentialsFramesetPath = FrameIO.CONTACTS_PATH + username + "-credentials" + File.separator;
		if (!new File(credentialsFramesetPath).exists()) {
			return null;
		}
		Scanner in = new Scanner(new File(credentialsFramesetPath + "credentials.inf"));
		String credentialsFrameNumber = in.nextLine().replace(ExpReader.EXTENTION, "");
		in.close();
		Frame frame = FrameIO.LoadFrame(username + "-credentials" + credentialsFrameNumber, FrameIO.CONTACTS_PATH);
		if (frame == null) {
			return null;
		}
		
		// obtain public key from frame
		Collection<Item> canditates = org.expeditee.auth.Actions.getByContent(frame, "PublicKey");
		String keyEncoded = "";
		for (Item i: canditates) {
			if (i.getData() != null) {
				keyEncoded = i.getData().get(0);
			}
		}
		if (keyEncoded.isEmpty()) {
			return null;
		}
		byte[] keyBytes = Base64.getDecoder().decode(keyEncoded);
		return KeyFactory.getInstance(AsymmetricAlgorithm).generatePublic(new X509EncodedKeySpec(keyBytes));
	}
	
	final void markRequestedColleagues(String username) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException {
		KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(new SecretKeySpec(TRUE, SymmetricAlgorithm));
		KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(KeyList.PersonalKey.get().getText().toCharArray());
		keyStore.setEntry(username + "colleaguesRequested", entry, entryPassword);
		keyStore.store(new FileOutputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME), Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
	}
	
	final void clearRequestedColleagues(String username) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException {
		KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(new SecretKeySpec(FALSE, SymmetricAlgorithm));
		KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(KeyList.PersonalKey.get().getText().toCharArray());
		keyStore.setEntry(username + "colleaguesRequested", entry, entryPassword);
		keyStore.store(new FileOutputStream(FrameIO.PARENT_FOLDER + KEYSTOREFILENAME), Constants.CREDENTIALS_KEYSTORE_PASSWORD.toCharArray());
	}
	
	final boolean hasRequestedColleagues(String username) throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableEntryException {
		String alias = username + "colleaguesRequested";
		if (!keyStore.containsAlias(alias)) {
			return false;
		} else {
			KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(KeyList.PersonalKey.get().getText().toCharArray());
			KeyStore.SecretKeyEntry entry = (SecretKeyEntry) keyStore.getEntry(alias, entryPassword);
			return Arrays.equals(entry.getSecretKey().getEncoded(), TRUE);
		}
	}
	
//	final void putColleagues(String username, String[] colleagues) throws KeyStoreException {
//		String alias = username + "colleagues";
//		final SecretKeySpec secretKeySpec = new SecretKeySpec((colleagues[0] + System.getProperty("line.separator") + colleagues[1]).getBytes(), SymmetricAlgorithm);
//		KeyStore.SecretKeyEntry entry = new KeyStore.SecretKeyEntry(secretKeySpec);
//		KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(KeyList.PersonalKey.get().getText().toCharArray());
//		keyStore.setEntry(alias, entry, entryPassword);
//	}
//	
//	final String[] getColleagues(String username) throws KeyStoreException, NoSuchAlgorithmException, UnrecoverableEntryException {
//		String alias = username + "colleagues";
//		if (!keyStore.containsAlias(alias)) {
//			return null;
//		} else {
//			KeyStore.ProtectionParameter entryPassword = new KeyStore.PasswordProtection(KeyList.PersonalKey.get().getText().toCharArray());
//			KeyStore.SecretKeyEntry entry = (SecretKeyEntry) keyStore.getEntry(alias, entryPassword);
//			byte[] colleaguesEncoded = entry.getSecretKey().getEncoded();
//			String colleagues = new String(colleaguesEncoded);
//			return colleagues.split(System.getProperty("line.separator"));
//		}
//	}
	
	private static void setInputManagerWindowRoutines()	{
		InputManager manager = EcosystemManager.getInputManager();
		
	    // Refresh the layout when the window resizes
		manager.addWindowEventListener(new WindowEventListener() {
			@Override
			public void onWindowEvent(WindowEventType type)
			{
				if (type != WindowEventType.WINDOW_RESIZED) {
					return;
				}
				DisplayController.refreshWindowSize();
				FrameIO.RefreshCacheImages();
				for (Frame frame : DisplayController.getFrames()) {
					if (frame != null) {
						ItemUtils.Justify(frame);
						frame.refreshSize();
					}
				}
				DisplayController.requestRefresh(false);
			}
		});
		
		manager.addWindowEventListener(new WindowEventListener() {
			@Override
			public void onWindowEvent(WindowEventType type)
			{
				if (type != WindowEventType.MOUSE_EXITED_WINDOW) {
					return;
				}
				StandardGestureActions.mouseExitedWindow();
			}
		});
		
		manager.addWindowEventListener(new WindowEventListener() {
			@Override
			public void onWindowEvent(WindowEventType type)
			{
				if (type != WindowEventType.MOUSE_ENTERED_WINDOW) {
					return;
				}
				StandardGestureActions.mouseEnteredWindow();
			}
		});
		
		manager.addWindowEventListener(new WindowEventListener() {
			@Override
			public void onWindowEvent(WindowEventType type)
			{
				if (type != WindowEventType.WINDOW_CLOSED) {
					return;
				}
				if (Browser._theBrowser != null) {
					Browser._theBrowser.exit();
				}
			}
		});
	}
}
