package org.expeditee.auth;

import java.io.IOException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.Base64;
import java.util.List;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.expeditee.io.ExpWriter;
import org.expeditee.items.Text;
import org.expeditee.settings.identity.secrets.KeyList;
import org.ngikm.cryptography.CryptographyConstants;

public class EncryptedExpWriter extends ExpWriter implements CryptographyConstants {
	private SecretKey key;
	private String label;
	private static final String nl = "\n";
	
	public EncryptedExpWriter(String encryptionLabel) throws IOException {
		if (encryptionLabel.compareTo("Profile") == 0) {
			// obtain personal key
			Text text = KeyList.PersonalKey.get();
			List<String> data = text.getData();
			if (data != null && !data.isEmpty()) {
				byte[] keyBytes = Base64.getDecoder().decode(data.get(0));
				key = new SecretKeySpec(keyBytes, SymmetricAlgorithm);
			}
			label = "Profile";
		} else {
			byte[] keyBytes = resolveKeyFromLabel(encryptionLabel);
			key = new SecretKeySpec(keyBytes, SymmetricAlgorithm);
			label = encryptionLabel;
		}
	}

	@Override
	protected void preOutputFrame() {
		try {
			String line = EncryptedExpReader.ENCRYPTED_EXP_FLAG + label + nl;
			_writer.write(line);
			_stringWriter.append(line);
		} catch (final IOException e) { 
			e.printStackTrace();
		}
	}
	
	@Override
	protected void writeLine(String line) throws IOException {
		// do not write empty lines
		if (line == null) { return; }
		
		// prepare line to write out
		byte[] encrypted = EncryptSymmetric(line.getBytes(), key);
		String toWrite = Base64.getEncoder().encodeToString(encrypted) + nl;
		
		// output
		_writer.write(toWrite);
		_stringWriter.append(toWrite);
	}
	
	private byte[] resolveKeyFromLabel(String label) {
		return null;
	}
	
	private static byte[] EncryptSymmetric(byte[] toEncrypt, SecretKey key) {
		try {
			Cipher cipher = Cipher.getInstance(SymmetricAlgorithm + SymmetricAlgorithmParameters);
			cipher.init(Cipher.ENCRYPT_MODE, key);
			//could use modulus
			int length = (int) ((Math.ceil(toEncrypt.length / 16f)) * 16);
			byte[] toEncryptSizeAdjusted = Arrays.copyOf(toEncrypt, length);
			byte[] result = cipher.doFinal(toEncryptSizeAdjusted);
			return result;
		} catch (final NoSuchAlgorithmException | NoSuchPaddingException | InvalidKeyException
				| IllegalBlockSizeException | BadPaddingException e) {
			e.printStackTrace();
			return null;
		}
	}
}
